#!/bin/sh
# mini shell chat
info="mini shell chat // 2023-04-02 yargo@zaibatsu.circumlunar.space"

# chat lines displayed after each command
tail=3
# chat lines displayed with .l command
last=10
# nick is user name or anonymous
nick=${USER:-anonymous}

# default and minimal linewidth and display height
defcols=77
defrows=23
mincols=16
minrows=4

cols=$defcols
rows=$defrows

# files: chat contains the entered text, bang the login/logout information
# (can be defined independently by environment variables)
# (do not rename/reformat! will be modified by install.sh)
chat=/tmp/schatfile.txt
bang=/tmp/schatbang.txt
# suffix for lockfiles
lksuffix=lock

# configuration file
conf=$HOME/.schatnrc

conftbeg='['
conftend=']'
confnbeg='<'
confnend='>'
conflend=''
filter=yes

if test -r "$conf"
# process all permitted configuration variables
# (chat and bang left out for security reasons,
# as they allow accessing any file; are normally
# hardcoded for a given host with install.sh)
then
 echo :: reading configuration file $conf
 for cv in ftbeg ftend fnbeg fnend flend filter rows cols
# read lines with current var, only take last one
# and replace backslash and separator (|) used below at --S--
do ln=`grep "^$cv " "$conf"|tail -n 1|tr '\\\\|' '/!'`
# get value (after separating SPC)
 val=${ln#* }
 case $ln in
 ftbeg*) conftbeg="$val" ;;
 ftend*) conftend="$val" ;;
 fnbeg*) confnbeg="$val" ;;
 fnend*) confnend="$val" ;;
 flend*) conflend="$val" ;;
 filter*) filter="$val" ;;
 chat*) chat="$val" ;;
 bang*) bang="$val" ;;
 rows*) rows="$val" ;;
 cols*) cols="$val" ;;
 esac
done
else
 echo :: initializing configuration file $conf
 cat <<EOS > "$conf"
# configuration file for $0
# automatically generated `date -u`
# *name-key-pairs must be separated by one SPC*
# string printed before time-stamp:
ftbeg $conftbeg
# string printed after time-stamp:
ftend $conftend
# string printed before nick:
fnbeg $confnbeg
# string printed after nick:
fnend $confnend
# string printed at end of line:
flend $conflend
# flag for filtering control characters, must be yes or no
filter $filter
# file with chat messages
#chat $chat
# file with chat enter/leave notifications
#bang $bang
# linewidth, columns
cols ${cols:-$defcols}
# display height, rows
rows ${rows:-$defrows}
EOS
fi

# override, if given as script arguments;
# 0 is same as unspecified, allows to keep
# cols from config file, while overriding rows
if test "$1" != "" -a "$1" != "0"
then cols=$1
fi
if test "$2" != "" -a "$2" != "0"
then rows=$2
fi

# display formatted time stamp and appended arguments
now() {
 echo "`date -u +%Y-%m-%d,%H:%MZ` $*"
}

# finish in an orderly way
cleanup() {
# log leaving chat
 now ::::: $nick leaves | appendto $bang
 echo
# final display of last chat lines
 fout "`tail -n $tail $chat`"
# kill jobs for continuous log display
 kill $bid $tpid
 echo ':: bye!'
}

# formatted output, uses global vars $cols and $rows
# arg1= text to output
fout() {
 local tx tl lw sh l1 l2
 lw=${cols:-$defcols}
 lw=`echo $lw | tr -c -d 0-9`
 if test ${lw:-0} -lt $mincols
 then lw=$mincols
 fi
 sh=${rows:-$defrows}
 sh=`echo $sh | tr -c -d 0-9`
 if test ${sh:-0} -lt $minrows
 then sh=$minrows
 fi
# apply filter if option is set (replace by SPC), replace line elements
# according to configuration settings, and fold; make sure separator(|)
# in sed pattern further down is same as above at --S--
 tx=`echo "$1" | { if test "$filter" = "yes"
  then tr -c '

 -~' ' '
  else cat
  fi
 } |
 sed -e "s|\([^ ]*\)  *\([^ ]*\)  *\(.*\)|$conftbeg\1$conftend $confnbeg\2$confnend \3$conflend|" |
 fold -s -w $lw`
# number of lines
 tl=`echo "$tx" | wc -l`
# page limits (line numbers)
 l1=1
 l2=$sh
# display pages until end reached
 while test $l1 -le $tl
 do
# display one page
  echo "$tx" | sed -n -e "${l1},${l2}p"
# wait for go-ahead, unless last page
  if test $l2 -lt $tl
  then
   echo '::: <RETURN> / <Q>uit :::'
   read lw
# exit if interrupted by trapped signal
   if test $? -ne 0
   then exit
   fi
   case $lw in
   q*|Q*) l1=$tl ; l2=$tl ;;
   esac
  fi
# shift page limits
  l1=$(( $l1+$sh ))
  l2=$(( $l2+$sh ))
 done
}

# check for a lockfile (used in other scripts managing chatfiles)
checklock() {
 local lf
 lf="$1.$lksuffix"
 if test -f "$lf"
 then
  echo :: $1 is locked -- please wait 5 sec
  sleep 5
  if test -f "$lf"
  then
   echo :: still locked -- please wait another 5 sec
   sleep 5
   if test -f "$lf"
   then
    echo :: still locked, lockfile contents:
    sed -e 's/^/::  /' "$lf"
    echo :: trying to ignore it now, please verify whether it worked
    sleep 1
   fi
  fi
 fi
}

# append text after lockcheck
appendto() {
 checklock "$1"
 cat >> "$1"
}

# display schat prompt
sprompt() {
 echo -n `now` ''
}

# normalize command string
normcz() {
 cz=`echo "$cz" | sed -e 's/^[!/.]/./'`
}

help() {
cat <<EOT

:: $info ::
::  you are $nick ::
:: dot-commands (lines beginning with '.' or '/' or '!'):
::  .q to stop, .l N to show last N (or $last if missing) lines
::  .h (or any undefined dot-command) for this help
::  .w to list other users currently running schat
::  .f for 'tail -f' on chatfile
::  .char to toggle filtering of control characters
::  .cols N and .rows N to set formatting values (see below)
::    (may be preset in configuration file)
:: all other lines will be appended to chatfile
::  (except for one-character entries, which probably are typos,
::  and exit and quit, which are assumed to be wishes for quitting)
:: after a line appended, the $tail most recent lines are displayed
:: script can be given two arguments for formatted output:
::  linewidth (cols, now $cols) for folding, and
::  display height (rows, now $rows) for paging
::   0=unspecified, i.e keep value from configuration file
::  (minimal values: cols>=$mincols, rows>=$minrows)
:: delimiter strings and linewidth/height for formatted output
:: can be defined in configuration file
::  $conf

EOT
}

# chat and log files at least must be readable
if test ! -r $chat
 then
 echo "$chat not readable!"
 exit 1
fi
if test ! -r $bang
 then
 echo "$bang not readable!"
 exit 1
fi

# log entering chat
now ::::: $nick enters | appendto $bang

help

# continuous display of log file
tail -f -n 5 $bang &
# save PID of display job
bid=$!

# catch signals issued by user (keyboard)
trap cleanup HUP INT QUIT ABRT TERM

# clear command/chat entry line
cz=''
while test "$cz" != ".q" -a "$cz" != ".quit"
do
# preset number of chat lines to be displayed
 l=$tail
 case $cz in
# process possible command (beginning with dot)
  .*) {
   case $cz in
# set formatting parameters
    .cols) cols=${a1:-$cols} ;;
    .rows) rows=${a1:-$rows} ;;
# .char: toggle filter flag
    .char) if test "$filter" = "yes"
     then filter=no
     else filter=yes
     fi
     echo :: filter for special characters = $filter
     ;;
# .l: set number of chat lines to be displayed
    .l) l=${a1:-$last} ;;
# .f: continuous monitoring of chat file with 'tail'
    .f) tail -f $chat &
     tpid=$!
     echo :: tailing chatfile until next entry
# wait and read new entry line
     sprompt
     read cz a1 az
# exit if interrupted by trapped signal
     if test $? -ne 0
     then exit
     fi
     normcz
# kill monitoring job
     kill $tpid
# clear PID for cleanup
     tpid=''
     continue
     ;;
# .w: get running jobs, filter for schat, get UIDs and translate to user names, sort and reduce to unique entries
    .w) ps -a -eo ruid,args | grep -v tail | grep [s]chat | { while read n _ ; do id -n -u $n ; done ; } | sed -e 's/^/: /' | sort -u ;;
# any other "command" results in help displayed
    .*) help ;;
   esac
   } ;;
# other lines are appended to chat file, with time stamp
  ??*) if test "$cz" = "exit" -o "$cz" = "quit"
# help clueless users who desperately type exit or quit
   then if test "$a1" = ""
    then cleanup
     exit
    fi
   fi
   now "$nick $cz $a1 $az" | appendto $chat
  ;;
# if first word is just one character, only append if more words present
  ?) if test "$a1" = ""
     then echo ':: one-character entry ignored!'
     else now "$nick $cz $a1 $az" | appendto $chat
     fi ;;
 esac
 echo
# display recent chat lines
 fout "`tail -n $l $chat`"
# get new entry line
 sprompt
 read cz a1 az
# exit if interrupted by trapped signal
 if test $? -ne 0
 then exit
 fi
 normcz
done

cleanup
